/* PSPP - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */


/* This #if encloses the rest of the file. */
#if !NO_HTML

#include <config.h>
#include <assert.h>
#include <errno.h>

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "error.h"		  
#include "output.h"
#include "som.h"
#include "filename.h"
#include "version.h"
#include "getline.h"

/* HTML output driver extension record. */
typedef struct
  {
    /* User parameters. */
    char *prologue_fn;		/* Prologue's filename relative to font dir. */

    /* Internal state. */
    file_ext file;		/* Output file. */
    int sequence_no;		/* Sequence number. */
  }
html_driver_ext;

/* Prototypes. */
static int postopen (file_ext *);
static int preclose (file_ext *);

int
html_open_global (outp_class *this)
{
  return 1;
}

int
html_close_global (outp_class *this)
{
  return 1;
}

int
html_init_driver (outp_driver *this, int outp_i)
{
  html_driver_ext *x;

  switch (outp_i)
    {
    case OUTP_I_INIT:
      {
	verbose_msg (1, _("HTML driver initializing as `%s'..."), this->name);

	this->ext = x = (html_driver_ext *) xmalloc (sizeof (html_driver_ext));
	this->res = 0;
	this->horiz = this->vert = 0;
	this->width = this->length = 0;

	x->prologue_fn = NULL;

	x->file.filename = NULL;
	x->file.mode = "w";
	x->file.file = NULL;
	x->file.sequence_no = &x->sequence_no;
	x->file.param = this;
	x->file.postopen = postopen;
	x->file.preclose = preclose;

	x->sequence_no = 0;
      }
      break;
    case OUTP_I_OPTIONS_INIT:
      {
	x = this->ext;
	
	if (NULL == x->file.filename)
	  x->file.filename = xstrdup (HTML_DEFAULT_OUTPUT_FILE);
	
	if (x->prologue_fn == NULL)
	  x->prologue_fn = xstrdup ("html-prologue");

	verbose_msg (1, _("%s: Initialization complete."), this->name);
      }
      break;
    case OUTP_I_DONE:
      {
	int i;

	verbose_msg (1, _("%s: Beginning closing..."), this->name);
	x = this->ext;
	close_file_ext (&x->file);
	free (x->prologue_fn);
	free (x->file.filename);
	free (x);
	verbose_msg (1, _("%s: Finished closing."), this->name);
      }
      break;
    }
  return 1;
}

/* Generic option types. */
enum
{
  boolean_arg = -10,
  string_arg,
  nonneg_int_arg,
};

/* All the options that the HTML driver supports. */
static outp_option option_tab[] =
{
  /* *INDENT-OFF* */
  {"output-file",		1},
  {"prologue-file",		string_arg,	0},
  {""},
  /* *INDENT-ON* */
};
static outp_option_info option_info;

void
html_option (outp_driver *this, char *key, outp_value *val)
{
  html_driver_ext *x = this->ext;
  int cat, subcat;
  a_string value;

  value = val->string;

  cat = outp_match_keyword (key, option_tab, &option_info, &subcat);
  if (val->type != 'a')
    {
      msg (SE, _("String value expected for key `%s'."), key);
      return;
    }

  switch (cat)
    {
    case 0:
      msg (SE, _("Unknown configuration parameter `%s' for HTML device "
	   "driver."), key);
      break;
    case 1:
      free (x->file.filename);
      x->file.filename = sa_strdup (&value);
      break;
    case string_arg:
      {
	char **dest;
	switch (subcat)
	  {
	  case 0:
	    dest = &x->prologue_fn;
	    break;
	  default:
	    assert (0);
	  }
	if (*dest)
	  free (*dest);
	*dest = sa_strdup (&value);
      }
      break;
#if __CHECKER__
    case 42000:
      assert (0);
#endif
    default:
      assert (0);
    }
}

/* Variables for the prologue. */
typedef struct
  {
    char *key;
    char *value;
  }
html_variable;
  
static html_variable *html_var_tab;

/* Searches html_var_tab for a html_variable with key KEY, and returns
   the associated value. */
static char *
html_get_var (char *key)
{
  html_variable *v;

  for (v = html_var_tab; v->key; v++)
    if (streq (key, v->key))
      return v->value;
  return NULL;
}

/* Writes the HTML prologue to file F. */
static int
postopen (file_ext *f)
{
  static html_variable dict[] =
    {
      {"generator", 0},
      {"date", 0},
      {"user", 0},
      {"host", 0},
      {"title", 0},
      {"subtitle", 0},
      {"source-file", 0},
      {0, 0},
    };
#if HAVE_UNISTD_H
  char host[128];
#endif
  time_t curtime;
  struct tm *loctime;
  char *title;
  int curfn_len;

  outp_driver *this = f->param;
  html_driver_ext *x = this->ext;

  char *prologue_fn = search_path (x->prologue_fn, config_path, NULL);
  FILE *prologue_file;

  char *buf = NULL;
  int buf_size = 0;

  if (prologue_fn == NULL)
    {
      msg (IE, _("Cannot find HTML prologue.  The use of `-vv' "
		 "on the command line is suggested as a debugging aid."));
      return 0;
    }

  verbose_msg (1, _("%s: Reading HTML prologue \"%s\"..."),
	       this->name, prologue_fn);
  prologue_file = fopen (prologue_fn, "rb");
  if (prologue_file == NULL)
    {
      fclose (prologue_file);
      free (prologue_fn);
      msg (IE, "%s: %s", prologue_fn, strerror (errno));
      goto error;
    }

  dict[0].value = version;

  curtime = time (NULL);
  loctime = localtime (&curtime);
  dict[1].value = asctime (loctime);
  {
    char *cp = strchr (dict[1].value, '\n');
    if (cp)
      *cp = 0;
  }

  /* PORTME: Determine username, net address. */
#if HAVE_UNISTD_H
  dict[2].value = getenv ("LOGNAME");
  if (!dict[2].value)
    dict[2].value = getlogin ();
  if (!dict[2].value)
    dict[2].value = _("nobody");

  if (gethostname (host, 128) == -1)
    if (errno == ENAMETOOLONG)
      host[127] = 0;
    else
      strcpy (host, _("nowhere"));
  dict[3].value = host;
#else /* !HAVE_UNISTD_H */
  dict[2].value = _("nobody");
  dict[3].value = _("nowhere");
#endif /* !HAVE_UNISTD_H */

  dict[4].value = outp_title ? outp_title : "";
  dict[5].value = outp_subtitle ? outp_subtitle : "";
  
  dict[6].value = curfn ? curfn : "<stdin>";

  html_var_tab = dict;
  while (-1 != getline (&buf, &buf_size, prologue_file))
    {
      char *cp;
      char *buf2;
      int len;

      if (strstr (buf, "!!!"))
	continue;
      
      {
	char *cp = strstr (buf, "!title");
	if (cp)
	  if (outp_title == NULL)
	    continue;
	  else
	    *cp = '\0';
      }
      
      {
	char *cp = strstr (buf, "!subtitle");
	if (cp)
	  if (outp_subtitle == NULL)
	    continue;
	  else
	    *cp = '\0';
      }
      
      /* PORTME: Line terminator. */
      buf2 = interp_vars (buf, html_get_var);
      len = strlen (buf2);
      fwrite (buf2, len, 1, f->file);
      if (buf2[len - 1] != '\n')
	putc ('\n', f->file);
      free (buf2);
    }
  if (ferror (f->file))
    msg (IE, _("Reading `%s': %s."), prologue_fn, strerror (errno));
  fclose (prologue_file);

  free (prologue_fn);
  free (buf);

  if (ferror (f->file))
    goto error;

  verbose_msg (1, _("%s: HTML prologue read successfully."), this->name);
  return 1;

error:
  verbose_msg (1, _("%s: Error reading HTML prologue."), this->name);
  return 0;
}

/* Writes the HTML epilogue to file F. */
static int
preclose (file_ext *f)
{
  outp_driver *this = f->param;
  html_driver_ext *x = this->ext;

  fprintf (f->file,
	   "</BODY>\n"
	   "</HTML>\n"
	   "<!-- end -->\n");

  if (ferror (f->file))
    return 0;
  return 1;
}

int
html_open_page (outp_driver *this)
{
  html_driver_ext *x = this->ext;
  
  x->sequence_no++;
  if (!open_file_ext (&x->file))
    {
      if (errno)
	/* FIXME: needs to be deleted from driver list */
	msg (ME, _("HTML output driver: %s: %s"), x->file.filename,
	     strerror (errno));
      return 0;
    }

  return !ferror (x->file.file);
}

int
html_close_page (outp_driver *this)
{
  html_driver_ext *x = this->ext;
  
  return !ferror (x->file.file);
}

void
html_submit (outp_driver *this, som_submission_form *f)
{
  html_driver_ext *x = this->ext;
  
  /* FIXME */
  if (x->sequence_no == 0 && !html_open_page (this))
    {
      msg (ME, _("Cannot open first page on HTML device %s."), this->name);
      return;
    }
  
  fflush (stdout);
  fprintf (x->file.file,
	   "<P>Table: %.*s</P>\n", f->title.l, f->title.s ? f->title.s : "");
}

/* HTML driver class. */
outp_class html_class =
{
  "html",
  0xfaeb,
  1,

  html_open_global,
  html_close_global,
  NULL,

  html_init_driver,
  html_option,

  html_open_page,
  html_close_page,

  html_submit,

  NULL,
  NULL,
  NULL,
  NULL,

  NULL,
  NULL,
  NULL,
  NULL,

  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
};

#endif /* !NO_HTML */

