/* PSPP - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

#include <config.h>
#include <ctype.h>
#include <assert.h>
#include <stdarg.h>
#include <limits.h>
#include "error.h"
#include "str.h"
#include "misc.h"
#include "arena.h"
#include "output.h"
#include "som.h"
#include "var.h"

/* This file contains all the som front-end routines. */

/* Declare virtual local variables. */
#define EXTERN extern
#include "somP.h"
#undef EXTERN

/* gcc complains if you pass a zero-length string to a printf-type
   function.  Using this string bypasses the warning. */
char zero_length[1] = "";

/* Related to output drivers. */

void
som_destroy_driver (outp_driver * d)
{
  d->class->close_page (d);
  free (d->som);
}

/* Tables. */

/* Pushes curtab_arena on the stack of table arenas.  (This stack
   allows one table to be output while another is being constructed,
   etc.) */
void
som_push_workspace (void)
{
  if (n_arena_stack >= m_arena_stack)
    {
      m_arena_stack += 8;
      arena_stack = xrealloc (arena_stack, sizeof (arena *) * m_arena_stack);
    }
  arena_stack[n_arena_stack++] = curtab_arena;
  curtab_arena = NULL;
}

/* Pops the TOS from arena_stack into curtab_arena. */
void
som_pop_workspace (void)
{
  assert (n_arena_stack);
  arena_destroy (&curtab_arena);
  curtab_arena = arena_stack[--n_arena_stack];
}

#if GLOBAL_DEBUGGING
void
som_check_workspace (void)
{
  if (n_arena_stack != 0)
    {
      msg (SE, _("som: Inconsistency in arena stack--height is %d."),
	   n_arena_stack);
      failure ();
    }
}
#endif

/* Creates a table with NC columns and NR rows according to
   CREATE_FLAGS. */
som_table *
som_create_table (int nc, int nr, int create_flags)
{
#if GLOBAL_DEBUGGING
  int i;
  som_any_cell **c;
#endif

  void *(*alloc_func) (arena **, size_t);

  som_table *t;

  assert (n_arena_stack &&
	  _("Please note that calls to som_push_workspace() and"
	  "som_pop_workspace() must be balanced around calls"
	  "to som_create_table()."));

  t = arena_alloc (&curtab_arena, sizeof (som_table));
  t->owner = NULL;
  t->x = t->y = -1;
  t->flags = STAB_COLUMNATE_NONE;
  t->nr = nr;
  t->nc = nc;
  t->l = t->r = t->t = t->b = 0;

  alloc_func = (create_flags & SOM_CREATE_ALLOW_RESIZE
		? arena_malloc
		: arena_alloc);

  t->c = alloc_func (&curtab_arena, nr * nc * sizeof (som_any_cell *));
  t->w = alloc_func (&curtab_arena, sizeof (int) * (2 * nc + 1));
  t->h = alloc_func (&curtab_arena, sizeof (int) * (2 * nr + 1));
  t->rh = alloc_func (&curtab_arena, nc * (nr + 1));
  memset (t->rh, 0, nc * (nr + 1));
  t->rv = alloc_func (&curtab_arena, (nc + 1) * nr);
  memset (t->rv, 0, (nc + 1) * nr);

#if GLOBAL_DEBUGGING
  i = nc * nr;
  c = t->c;
  while (i--)
    *c++ = NULL;
#endif

  return t;
}

/* Sets the size of a table, in rows.  This should only be use to
   _reduce_ the number of rows in a table, although once you've
   reduced that value, you can increase it back up to the original
   value.  Does *not* change amount of allocated memory. */
void
som_set_table_height (som_table * t, int nr)
{
  t->nr = nr;
}

/* Changes the size of a table.  This can be used to change either or
   both dimensions to larger or smaller values.  Consider use of the
   above routine first if it can be done efficiently.  If you're
   dealing with a _very_ large table (like the output of LIST), split
   it using the SOM_TOPT_PARTIAL_* interface instead.

   Changing the number of columns in a table is particularly expensive
   in space and time.  Avoid doing such.  In fact, transferring of
   rules isn't even implemented yet. */
void
som_change_table_size (som_table * t, int nc, int nr)
{
  int prev_nc = t->nc;
  int prev_nr = t->nr;

  if (nc != t->nc)
    {
      int mr1 = min (nr, prev_nr);
      int mc1 = min (nc, prev_nc);
#if DEBUGGING
      int mr2 = max (nr, prev_nr);
      int mc2 = max (nc, prev_nc);
#endif
      som_any_cell **new_c;
      som_any_cell **ncp;
      int c, r;

      new_c = arena_malloc (&curtab_arena, nr * nc * sizeof (som_any_cell *));
      ncp = new_c;
      for (r = 0; r < mr1; r++)
	{
	  for (c = 0; c < mc1; c++)
	    *ncp++ = t->c[c + r * t->nc];
#if DEBUGGING
	  for (; c < mc2; c++)
	    *ncp++ = NULL;
#else
	  ncp += abs (nc - prev_nc);
#endif
	}
#if DEBUGGING
      for (; r < mr2; r++)
	for (c = 0; c < mc2; c++)
	  *ncp++ = NULL;
#endif
      arena_free (t->c);
      t->c = new_c;

      t->w = arena_realloc (&curtab_arena, t->w, sizeof (int) * (2 * nc + 1));

#if 0
      /* Not fully implemented. */
      t->rh = arena_realloc (&curtab_arena, t->rh, nc * (nr + 1));
#endif
    }
  else if (nr != t->nr)
    {
      t->c = arena_realloc (&curtab_arena, t->c,
			    nr * nc * sizeof (som_any_cell *));
#if DEBUGGING
      if (nr > t->nr)
	{
	  int r, c;

	  for (r = t->nr; r < nr; r++)
	    for (c = 0; c < nc; c++)
	      t->c[c + r * nc] = NULL;
	}
#endif

      t->rh = arena_realloc (&curtab_arena, t->rh, nc * (nr + 1));
      t->rv = arena_realloc (&curtab_arena, t->rv, (nc + 1) * nr);
      if (nr > t->nr)
	{
	  memset (&t->rh[nc * (t->nr + 1)], 0, (nr - t->nr) * nc);
	  memset (&t->rv[(nc + 1) * t->nr], 0, (nr - t->nr) * (nc + 1));
	}
    }
  else
    return;

  t->h = arena_realloc (&curtab_arena, t->h, sizeof (int) * (2 * nr + 1));

  t->nr = nr;
  t->nc = nc;
}


/* Sets the number of header rows on each side of TABLE to L on the
   left, R on the right, T on the top, B on the bottom.  Header rows
   are special because they are repeated when a table is broken across
   multiple columns or multiple pages. */
void
som_set_headers (som_table * table, int l, int r, int t, int b)
{
  table->l = l;
  table->r = r;
  table->t = t;
  table->b = b;
}

/* Draws a vertical line to the left of cells at horizontal position X
   from Y1 to Y2 inclusive in style STYLE, if style is not -1. */
void
som_vline (som_table * t, int style, int x, int y1, int y2)
{
  int y;

#if GLOBAL_DEBUGGING
  if (x < 0 || x > t->nc || y1 < 0 || y1 >= t->nr || y2 < 0 || y2 >= t->nr)
    {
      printf (_("bad vline: x=%d y=(%d,%d) in table size (%d,%d)\n"),
	      x, y1, y2, t->nc, t->nr);
      return;
    }
#endif
  if (style != -1)
    for (y = y1; y <= y2; y++)
      t->rv[x + (t->nc + 1) * y] = style;
}

/* Draws a horizontal line above cells at vertical position Y from X1
   to X2 inclusive in style STYLE, if style is not -1. */
void
som_hline (som_table * t, int style, int x1, int x2, int y)
{
  int x;

#if GLOBAL_DEBUGGING
  if (x1 < 0 || x1 >= t->nc || x2 < 0 || x2 >= t->nc || y < 0 || y > t->nr)
    {
      printf (_("bad hline: x=(%d,%d) y=%d in table size (%d,%d)\n"),
	      x1, x2, y, t->nc, t->nr);
      return;
    }
#endif
  if (style != -1)
    for (x = x1; x <= x2; x++)
      t->rh[x + t->nc * y] = style;
}

/* Draws a box around cells (X1,Y1)-(X2,Y2) inclusive with style F_H
   for horizontal lines and style F_V for vertical lines.  Fills the
   interior of the box with lines of style I_H for horizontal lines
   and styles I_V for vertical lines.  Any of the line styles may be
   -1 to avoid drawing those lines.  This is distinct from 0, which
   draws a null line. */
void
som_box (som_table * t, int f_h, int f_v, int i_h, int i_v,
	 int x1, int y1, int x2, int y2)
{
  int x, y;

#if GLOBAL_DEBUGGING
  if (x1 < 0 || x1 >= t->nc || x2 < 0 || x2 >= t->nc
      || y1 < 0 || y1 >= t->nr || y2 < 0 || y2 >= t->nr)
    {
      printf (_("bad box: (%d,%d)-(%d,%d) in table size (%d,%d)\n"),
	      x1, y1, x2, y2, t->nc, t->nr);
      return;
    }
#endif

  if (f_h != -1)
    for (x = x1; x <= x2; x++)
      {
	t->rh[x + t->nc * y1] = f_h;
	t->rh[x + t->nc * (y2 + 1)] = f_h;
      }
  if (f_v != -1)
    for (y = y1; y <= y2; y++)
      {
	t->rv[x1 + (t->nc + 1) * y] = f_v;
	t->rv[(x2 + 1) + (t->nc + 1) * y] = f_v;
      }

  if (i_h != -1)
    for (y = y1 + 1; y <= y2; y++)
      for (x = x1; x <= x2; x++)
	t->rh[x + t->nc * y] = i_h;
  if (i_v != -1)
    for (x = x1 + 1; x <= x2; x++)
      for (y = y1; y <= y2; y++)
	t->rv[x + (t->nc + 1) * y] = i_v;
}

/* `Columnates' table T, that is, sets it up so that, if it is narrow
   enough, it will appear in columns across the page instead of taking
   up lots of vertical space.

   If ACROSS is nonzero, rows will read left to right, then top to
   bottom.  If ACROSS is zero, rows will read top to bottom, then left
   to right, like reading columns in a newspaper.  GROUP is the number
   of rows to take as a unit.

   Currently columnated tables must be toplevel tables. */
void
som_columnate (som_table * t, int across, int group)
{
  if (t->flags != STAB_COLUMNATE_NONE)
    return;
  if (across)
    t->flags |= STAB_COLUMNATE_ACRS;
  else
    t->flags |= STAB_COLUMNATE_DOWN;
  t->group = group;
}

/* Returns the number of bytes occupied by cell C. */
static int
cell_byte_size (som_any_cell * c)
{
  assert (c);
  switch (c->any.con)
    {
    case SCON_EMPTY:
      return sizeof (som_empty_cell);
    case SCON_VALUE:
    case SCON_TEXT:
      return sizeof (som_text_cell) + LOAD_2 (&c->text.s.l1) - 1;
    case SCON_TABLE:
      return sizeof (som_table_cell);
    case SCON_GRID:
      return sizeof (som_grid_cell) + (c->grid.nc * c->grid.nr) - 1;
    case SCON_DRAW:
      assert (0);
    case SCON_JOIN_AUX:
      return sizeof (som_join_aux_cell);
    default:
      assert (0);
    }
#if __GNUC__
  abort ();
#endif
}

/* Joins the cells in the rectangle (X1,Y1)-(X2,Y2) in table T to form
   a single cell.  Row Y2 and column X2 are not included in the
   rectangle.  All cells in the rectangle other than (X1,Y1) must be
   empty; (X1,Y1) must *not* be empty.  (X1,Y1)-(X2,Y2) is not
   ranged-checked. */
void
som_join_cells (som_table * t, int x1, int y1, int x2, int y2)
{
  int x, y;

  for (x = x1; x < x2; x++)
    for (y = y1; y < y2; y++)
      if (x != x1 || y != y1)
	{
	  som_join_aux_cell *c;
	  assert (t->c[x + y * t->nc] == NULL);

	  c = arena_alloc (&curtab_arena, sizeof (som_join_aux_cell));
	  t->c[x + y * t->nc] = (som_any_cell *) c;

	  c->opt = SOPT_X_NONE;
	  c->con = SCON_JOIN_AUX;
	  c->c = x1;
	  c->r = y1;
	}
      else
	{
	  /* Copy the current cell to a new cell, tacking a
	     som_join_header on the front end.

	     This is a crock.  */
	  som_any_cell *c;
	  som_join_header *cn;
	  int nbytes;

	  c = t->c[x + y * t->nc];
	  assert (c);
	  nbytes = cell_byte_size (c);

	  c->any.opt = ((t->c[x + y * t->nc]->any.opt & ~SOPT_X_MASK)
			| SOPT_X_JOIN);

	  cn = arena_alloc (&curtab_arena, sizeof (som_join_header) + nbytes);
	  memcpy (&((char *) cn)[sizeof (som_join_header)], c, nbytes);
	  cn->d.nc = x2 - x1;
	  cn->d.nr = y2 - y1;
	  t->c[x + y * t->nc] = (som_any_cell *) cn;
	}
}

/* Puts the table CHILD as a cell in PARENT at (C,R) with options OPT.
   The destination cell must be empty. */
void
som_insert_table (som_table * parent, int c, int r, som_table * child,
		  unsigned char opt)
{
  som_table_cell *cell;

  assert (child && parent && child->owner == NULL
	  && parent->c[c + r * parent->nc] == NULL);
  child->owner = parent;
  child->x = c;
  child->y = r;

  cell = arena_alloc (&curtab_arena, sizeof (som_table_cell));
  parent->c[c + r * parent->nc] = (som_any_cell *) cell;

  /* FIXME?  SOPT_X_* other than SOPT_X_NORM are generally undesirable
     in the current setup. */
  cell->opt = opt & ~SOPT_X_MASK;
  cell->con = SCON_TABLE;
  cell->t = child;
}

/* Cells. */

/* Sets cell (C,R) in TABLE as an empty cell. */
void
som_set_null (som_table * table, int c, int r)
{
  som_empty_cell *cell;

  assert (table && c >= 0 && r >= 0 && c < table->nc && r < table->nr);

  cell = arena_alloc (&curtab_arena, sizeof (som_empty_cell));
  cell->opt = SOPT_NONE;
  cell->con = SCON_EMPTY;

  table->c[c + r * table->nc] = (som_any_cell *) cell;
}

/* Sets cell (C,R) in TABLE, with options OPT, to have COUNT lines of
   values taken from a vector at V, each displayed with the single
   format spec *F.  COUNT must be positive. */
void
som_set_value (som_table * table, int c, int r, unsigned char opt,
	       value * v, fmt_spec * f, int count)
{
  som_text_cell *cell;
  value temp_val;

  assert (table);
#if GLOBAL_DEBUGGING
  if (c < 0 || r < 0 || c >= table->nc || r >= table->nr)
    {
      printf (_("som_set_value(): bad cell (%d,%d) in table size (%d,%d)\n"),
	      c, r, table->nc, table->nr);
      return;
    }
#endif

  assert (count == 1);
  cell = arena_alloc (&curtab_arena, sizeof (som_text_cell) + f->w - 1);
  table->c[c + r * table->nc] = (som_any_cell *) cell;

  cell->opt = opt;
  cell->con = SCON_VALUE;
  STORE_2 (&cell->s.l1, f->w);
  if (formats[f->type].cat & FCAT_STRING)
    {
      temp_val.c = v->s;
      v = &temp_val;
    }
  convert_format_to_string (cell->s.s, f, v);
}

/* Sets cell (C,R) in TABLE, with options OPT, to have value VAL
   with NDEC decimal places. */
void
som_set_float (som_table * table, int c, int r, unsigned char opt,
	       double val, int ndec)
{
  som_text_cell *cell;
  char buf[40], *cp;
  
  fmt_spec f;
  
  f.type = FMT_F;
  f.w = 40;
  f.d = ndec;
  
  assert (table);
#if GLOBAL_DEBUGGING
  if (c < 0 || r < 0 || c >= table->nc || r >= table->nr)
    {
      printf (_("som_set_float(): bad cell (%d,%d) in table size (%d,%d)\n"),
	      c, r, table->nc, table->nr);
      return;
    }
#endif

  convert_format_to_string (buf, &f, (value *) & val);
  cp = buf;
  while (isspace (*cp) && cp < &buf[40])
    cp++;
  f.w = 40 - (cp - buf);

  cell = arena_alloc (&curtab_arena, sizeof (som_grid_cell) + f.w - 1);
  table->c[c + r * table->nc] = (som_any_cell *) cell;

  cell->opt = opt;
  cell->con = SCON_VALUE;
  STORE_2 (&cell->s.l1, f.w);
  memcpy (cell->s.s, cp, f.w);
}

/* Sets cell (C,R) in TABLE, with options OPT, to have text value TEXT. */
/* FIXME: Implement SOT_FIX. */
void
som_set_text (som_table * table, int c, int r, unsigned opt,
	      char *buf,...)
{
  va_list args;
  som_text_cell *cell;

  char temp_buf[1024];
  char *text;
  int len;

  assert (table);
#if GLOBAL_DEBUGGING
  if (c < 0 || r < 0 || c >= table->nc || r >= table->nr)
    {
      printf (_("som_set_text(): bad cell (%d,%d) in table size (%d,%d)\n"),
	      c, r, table->nc, table->nr);
      return;
    }
#endif

  if (opt & SOT_PRINTF)
    {
      va_start (args, buf);
      text = temp_buf;
      len = nvsprintf (text, buf, args);
      va_end (args);
    }
  else
    {
      text = buf;
      len = strlen (text);
    }

  if ((opt & SOT_TITLE) == SOT_TITLE)
    {
      cell = arena_alloc (&curtab_arena, sizeof (som_text_cell) + len - 1 + 5);
      table->c[c + r * table->nc] = (som_any_cell *) cell;

      cell->opt = opt;
      cell->con = SCON_TEXT;

      STORE_2 (&cell->s.l1, len + 5);
      cell->s.s[0] = 0;
      STORE_2 (&cell->s.s[1], 5);
      cell->s.s[3] = TAG_FONT_BY_POSITION;
      cell->s.s[4] = OUTP_F_BI;
      memcpy (&cell->s.s[5], text, len);
    }
  else
    {
      cell = arena_alloc (&curtab_arena, sizeof (som_text_cell) + len - 1);
      table->c[c + r * table->nc] = (som_any_cell *) cell;

      cell->opt = opt;
      cell->con = (opt & SOT_RICH ? SCON_TEXT : SCON_VALUE);

      STORE_2 (&cell->s.l1, len);
      memcpy (cell->s.s, text, len);
    }
}

/* Miscellaneous. */

/* Outputs text STRING as a table with a single cell having cell
   options OPTIONS, which is a combination of the SOPT_ and SOT_*
   constants. */
void
som_output_text (int options, char *string,...)
{
  som_submission_form f =
  {SOM_SUB_ORDINARY,
   SOM_TOPT_HEADER | SOM_TOPT_SPACING, NULL,
   {0, NULL}};
  char *buf = local_alloc (4096);
  char *text;

  som_push_workspace ();
  f.t = som_create_table (1, 1, SOM_CREATE_NONE);

  if (options & SOT_PRINTF)
    {
      va_list args;

      va_start (args, string);
      vsprintf (buf, string, args);
      va_end (args);
      text = buf;

      options &= ~SOT_PRINTF;
    }
  else
    text = string;

  som_set_text (f.t, 0, 0, options, text);
  som_submit_table (&f);

  local_free (buf);
  som_pop_workspace ();
}
