/* PSPP - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include "var.h"
#include "error.h"
#include "misc.h"
#include "output.h"
#include "som.h"
#include "sfm.h"
#include "tokens.h"
#include "avl.h"

/* Constants for DISPLAY utility. */
enum
  {
    AS_NAMES = 0,
    AS_INDEX,
    AS_VARIABLES,
    AS_LABELS,
    AS_DICTIONARY,
    AS_SCRATCH,
    AS_VECTOR,
  };

int describe_variable (variable *v, som_table *t, int r, int as);
     
/* SYSFILE INFO utility. */
int
cmd_sysfile_info (void)
{
  file_handle *h;
  dictionary *d;
  som_table *t;
  sfm_read_info inf;
  som_submission_form f;
  int ret_val = 1;
  int r, nr;
  int i;

  match_id (SYSFILE);
  match_id (INFO);

  match_id (FILE);
  match_tok ('=');

  h = fh_parse_file_handle ();
  if (!h)
    return 0;

  if (token != '.')
    ret_val = -3;

  d = sfm_read_dictionary (h, &inf);
  fh_close_handle (h);
  if (!d)
    return 0;

  som_push_workspace ();
  t = som_create_table (2, 9, SOM_CREATE_NONE);
  som_vline (t, SLIN_1 | SLIN_SPACING, 1, 0, 8);
  som_set_text (t, 0, 0, SOPT_A_LEFT, _("File:"));
  som_set_text (t, 1, 0, SOPT_A_LEFT | SOPT_X_HLTL, fh_handle_filename (h));
  som_set_text (t, 0, 1, SOPT_A_LEFT, _("Label:"));
  som_set_text (t, 1, 1, SOPT_A_LEFT | SOPT_X_HLTL,
		d->label ? d->label : _("No label."));
  som_set_text (t, 0, 2, SOPT_A_LEFT, _("Created:"));
  som_set_text (t, 1, 2, SOPT_A_LEFT | SOPT_X_HLTL | SOT_PRINTF, "%s %s by %s",
		inf.creation_date, inf.creation_time, inf.product);
  som_set_text (t, 0, 3, SOPT_A_LEFT, _("Endian:"));
  som_set_text (t, 1, 3, SOPT_A_LEFT | SOPT_X_HLTL,
		(inf.endianness == BIG ? _("Big.")
		 : (inf.endianness == LITTLE ? _("Little.")
		    : (assert (0), _("<internal error>")))));
  som_set_text (t, 0, 4, SOPT_A_LEFT, _("Variables:"));
  som_set_text (t, 1, 4, SOPT_A_LEFT | SOPT_X_HLTL | SOT_PRINTF, "%d",
		d->nvar);
  som_set_text (t, 0, 5, SOPT_A_LEFT, _("Cases:"));
  som_set_text (t, 1, 5, SOPT_A_LEFT | SOPT_X_HLTL | SOT_PRINTF,
		inf.ncases == -1 ? _("Unknown") : "%d", inf.ncases);
  som_set_text (t, 0, 6, SOPT_A_LEFT, _("Type:"));
  som_set_text (t, 1, 6, SOPT_A_LEFT | SOPT_X_HLTL, _("System File."));
  som_set_text (t, 0, 7, SOPT_A_LEFT, _("Weight:"));
  som_set_text (t, 1, 7, SOPT_A_LEFT | SOPT_X_HLTL,
		d->weight_var[0] ? d->weight_var : _("Not weighted."));
  som_set_text (t, 0, 8, SOPT_A_LEFT, _("Mode:"));
  som_set_text (t, 1, 8, SOPT_A_LEFT | SOPT_X_HLTL | SOT_PRINTF,
		_("Compression %s."), inf.compressed ? _("on") : _("off"));
  f.type = SOM_SUB_ORDINARY;
  f.options = SOM_TOPT_NONE;
  f.t = t;
  f.title.l = 0;
  f.title.s = NULL;
  som_submit_table (&f);
  som_pop_workspace ();

  som_push_workspace ();
  nr = 1 + 2 * d->nvar;
  t = som_create_table (3, nr, SOM_CREATE_ALLOW_RESIZE);
  som_set_headers (t, 0, 0, 1, 0);
  som_set_text (t, 0, 0, SOPT_A_LEFT | SOT_TITLE, _("Variable"));
  som_set_text (t, 1, 0, SOPT_A_LEFT | SOT_TITLE, _("Description"));
  som_set_text (t, 2, 0, SOPT_A_LEFT | SOT_TITLE, _("Position"));
  som_hline (t, SLIN_2, 0, 2, 1);
  for (r = 1, i = 0; i < d->nvar; i++)
    {
      if (r + 10 > nr)
	{
	  nr = max (nr * d->nvar / (i + 1), nr);
	  nr += 10;
	  som_change_table_size (t, 3, nr);
	}

      r = describe_variable (d->var[i], t, r, AS_DICTIONARY);
    }
  som_box (t, SLIN_1, SLIN_1, -1, SLIN_1, 0, 0, 2, r);
  som_set_table_height (t, r);

  f.type = SOM_SUB_ORDINARY;
  f.options = SOM_TOPT_HEADER | SOM_TOPT_SPACING;
  f.t = t;
  f.title.l = 0;
  f.title.s = NULL;
  som_submit_table (&f);
  som_pop_workspace ();

  free_dictionary (d);
  
  return ret_val;
}

/* DISPLAY utility. */

static void display_macros (void);
static void display_documents (void);
static void display_variables (variable **, int, int);

static int cmp_var_by_name (const void *, const void *);

int
cmd_display (void)
{
  /* Whether to sort the list of variables alphabetically. */
  int sorted;

  /* Variables to display. */
  int n;
  variable **vl;

  match_id (DISPLAY);

  if (match_id (MACROS))
    display_macros ();
  else if (match_id (DOCUMENTS))
    display_documents ();
  else if (match_id (FILE))
    {
      blank_line ();
      force_match_id (LABEL);
      if (default_dict.label == NULL)
	som_output_text (SOPT_A_LEFT,
			 _("The active file does not have a file label."));
      else
	{
	  som_output_text (SOPT_A_LEFT | SOT_TITLE, _("File label:"));
	  som_output_text (SOPT_A_LEFT | SOT_FIX, default_dict.label);
	}
    }
  else
    {
      static char *sbc[] =
      {"NAMES", "INDEX", "VARIABLES", "LABELS",
       "DICTIONARY", "SCRATCH", "VECTOR", NULL};
      char **cp;
      int as;

      sorted = match_id (SORTED);

      for (cp = sbc; *cp; cp++)
	if (token == ID && id_match (*cp, tokstr))
	  {
	    get_token ();
	    break;
	  }
      as = cp - sbc;

      if (*cp == NULL)
	as = AS_NAMES;

      match_tok ('/');
      match_id (VARIABLES);
      match_tok ('=');

      if (token != '.')
	{
	  if (!parse_variables (NULL, &vl, &n, PV_NONE))
	    {
	      free (vl);
	      return 0;
	    }
	  as = AS_DICTIONARY;	/* FIXME? */
	  printf ("n=%d\n", n);
	}
      else
	fill_all_vars (&vl, &n, FV_NONE);

      if (as == AS_SCRATCH)
	{
	  int i, m;
	  for (i = 0, m = n; i < n; i++)
	    if (vl[i]->name[0] != '#')
	      {
		vl[i] = NULL;
		m--;
	      }
	  n = m;
	}

      if (sorted)
	qsort (vl, n, sizeof (variable *), cmp_var_by_name);

      display_variables (vl, n, cp - sbc);

      free (vl);
    }

  if (token != '.')
    return msg (SE, _("A command terminator was expected at end of command."));
  return 1;
}

static int
cmp_var_by_name (const void *a, const void *b)
{
  return strcmp ((*((variable **) a))->name, (*((variable **) b))->name);
}

static void
display_macros (void)
{
  blank_line ();
  som_output_text (SOPT_A_LEFT, _("Macros not supported."));
}

static void
display_documents (void)
{
  blank_line ();
  if (default_dict.n_documents == 0)
    som_output_text (SOPT_A_LEFT, _("The active file dictionary does not "
		     "contain any documents."));
  else
    {
      char buf[81];
      int i;

      som_output_text (SOPT_A_LEFT | SOT_TITLE,
		       _("Documents in the active file:"));
      blank_line ();
      buf[80] = 0;
      for (i = 0; i < default_dict.n_documents; i++)
	{
	  int len = 79;

	  memcpy (buf, &default_dict.documents[i * 80], 80);
	  while ((isspace (buf[len]) || buf[len] == 0) && len > 0)
	    len--;
	  buf[len + 1] = 0;
	  som_output_text (SOPT_A_LEFT | SOT_FIX, buf);
	}
    }
}

static void
display_variables (variable **vl, int n, int as)
{
  variable **vp = vl;		/* Variable pointer. */
  som_table *t;
  int nc;			/* Number of columns. */
  int nr;			/* Number of rows. */
  int pc;			/* `Position column' */
  int r;			/* Current row. */
  som_submission_form f;
  int i;
  
  switch (as)
    {
    case AS_INDEX:
      nc = 2;
      break;
    case AS_NAMES:
      nc = 1;
      break;
    default:
      nc = 3;
      break;
    }

  som_push_workspace ();
  t = som_create_table (nc, n + 5, SOM_CREATE_ALLOW_RESIZE);
  nr = n + 5;
  som_hline (t, SLIN_2, 0, nc - 1, 1);
  som_set_text (t, 0, 0, SOPT_A_LEFT | SOT_TITLE, _("Name"));
  if (as != AS_NAMES)
    {
      pc = (as == AS_INDEX ? 1 : 2);
      som_set_text (t, pc, 0, SOPT_A_LEFT | SOT_TITLE, _("Position"));
    }
  if (as == AS_DICTIONARY || as == AS_VARIABLES)
    som_set_text (t, 1, 0, SOPT_A_LEFT | SOT_TITLE, _("Description"));

  for (i = r = 1; i <= n; i++)
    {
      variable *v;

      while (*vp == NULL)
	vp++;
      v = *vp++;

      if (as == AS_DICTIONARY || as == AS_VARIABLES)
	{
	  if (r + 10 > nr)
	    {
	      nr = max (nr * n / (i + 1), nr);
	      nr += 10;
	      som_change_table_size (t, 3, nr);
	    }

	  r = describe_variable (v, t, r, as);
	} else {
	  som_set_text (t, 0, r, SOPT_A_LEFT, v->name);
	  if (as != AS_NAMES)
	    som_set_text (t, pc, r, SOT_PRINTF, "%d", v->index + 1);
	  if (as == AS_DICTIONARY || as == AS_VARIABLES)
	    som_set_null (t, 1, r);
	  som_hline (t, SLIN_1, 0, 2, ++r);
	}
    }
  som_hline (t, SLIN_2, 0, 2, 1);
  som_box (t, SLIN_1, SLIN_1, -1, SLIN_1, 0, 0, 2, r);
  som_set_table_height (t, r);
  
  f.type = SOM_SUB_COLUMNATE_DOWN;
  f.options = SOM_TOPT_NONE;
  f.t = t;
  f.title.l = 0;
  f.title.s = NULL;
  som_submit_table (&f);
  som_pop_workspace ();
}

/* Puts a description of variable V into table T starting at row R.
   The variable will be described in the format AS.  Returns the next
   row available for use in the table. */
int 
describe_variable (variable *v, som_table *t, int r, int as)
{
  int prev_r = r;

  /* Put the name, var label, and position into the first row. */
  som_set_text (t, 0, r, SOPT_A_LEFT, v->name);
  som_set_text (t, 2, r, SOT_PRINTF, "%d", v->index + 1);

  if (as == AS_DICTIONARY)
    {
      som_set_text (t, 1, r, SOPT_A_LEFT | SOPT_X_HLTL,
		    v->label ? v->label : _("(no label)"));
      r++;
    }
  
  /* Print/write format, or print and write formats. */
  if (v->print.type == v->write.type
      && v->print.w == v->write.w
      && v->print.d == v->write.d)
    som_set_text (t, 1, r++, SOPT_A_LEFT | SOT_PRINTF, _("Format: %s"),
		  fmt_to_string (&v->print));
  else
    {
      som_set_text (t, 1, r++, SOPT_A_LEFT | SOT_PRINTF, _("Print Format: %s"),
		    fmt_to_string (&v->print));
      som_set_text (t, 1, r++, SOPT_A_LEFT | SOT_PRINTF, _("Write Format: %s"),
		    fmt_to_string (&v->write));
    }

  /* Missing values if any. */
  if (v->miss_type != MISSING_NONE)
    {
      char buf[80];
      char *cp = stpcpy (buf, _("Missing Values: "));

      if (v->type == NUMERIC)
	switch (v->miss_type)
	  {
	  case MISSING_1:
	    sprintf (cp, "%g", v->missing[0].f);
	    break;
	  case MISSING_2:
	    sprintf (cp, "%g; %g", v->missing[0].f, v->missing[1].f);
	    break;
	  case MISSING_3:
	    sprintf (cp, "%g; %g; %g", v->missing[0].f,
		     v->missing[1].f, v->missing[2].f);
	    break;
	  case MISSING_RANGE:
	    sprintf (cp, "%g THRU %g", v->missing[0].f, v->missing[1].f);
	    break;
	  case MISSING_LOW:
	    sprintf (cp, "LOWEST THRU %g", v->missing[0].f);
	    break;
	  case MISSING_HIGH:
	    sprintf (cp, "%g THRU HIGHEST", v->missing[0].f);
	    break;
	  case MISSING_RANGE_1:
	    sprintf (cp, "%g THRU %g; %g",
		     v->missing[0].f, v->missing[1].f, v->missing[2].f);
	    break;
	  case MISSING_LOW_1:
	    sprintf (cp, "LOWEST THRU %g; %g",
		     v->missing[0].f, v->missing[1].f);
	    break;
	  case MISSING_HIGH_1:
	    sprintf (cp, "%g THRU HIGHEST; %g",
		     v->missing[0].f, v->missing[1].f);
	    break;
	  default:
	    assert (0);
	  }
      else
	{
	  int i;

	  for (i = 0; i < v->miss_type; i++)
	    {
	      if (i != 0)
		cp = stpcpy (cp, "; ");
	      *cp++ = '"';
	      memcpy (cp, v->missing[i].s, v->width);
	      cp += v->width;
	      *cp++ = '"';
	    }
	  *cp = 0;
	}

      som_set_text (t, 1, r++,
		    SOPT_A_LEFT | (v->type == ALPHA ? SOT_RICH : 0), buf);
    }

  /* Value labels. */
  if (as == AS_DICTIONARY && v->val_lab)
    {
      som_table *sub;
      avl_traverser *trav = NULL;
      value_label *vl;
      int nvl = avl_count (v->val_lab);
      int i;

      /* FIXME: big nested tables. */
      if (nvl > 10)
	nvl = 10;

      sub = som_create_table (2, 1 + nvl, SOM_CREATE_NONE);
      som_box (sub, SLIN_1, SLIN_1, -1, SLIN_1, 0, 0, 1, nvl);
      som_hline (sub, SLIN_2, 0, 1, 1);
      som_set_headers (sub, 0, 0, 1, 0);

      som_set_text (sub, 0, 0, SOPT_A_LEFT, _("Value"));
      som_set_text (sub, 1, 0, SOPT_A_LEFT, _("Label"));
      for (i = 1, vl = avl_traverse (v->val_lab, &trav); vl;
	   i++, vl = avl_traverse (v->val_lab, &trav))
	{
	  char buf[128];

	  if (v->type == ALPHA)
	    {
	      memcpy (buf, vl->v.s, v->width);
	      buf[v->width] = 0;
	    }
	  else
	    sprintf (buf, "%g", vl->v.f);

	  som_set_text (sub, 0, i, SOPT_NONE, buf);
	  som_set_text (sub, 1, i, SOPT_A_LEFT | SOPT_X_BOTH, vl->s);

	  if (i == nvl) 
	    break;
	}

      for (;;)
	{
	  if (vl == NULL)
	    break;
	  vl = avl_traverse (v->val_lab, &trav);
	}

      som_insert_table (t, 1, r++, sub, SOPT_NONE);
    }

  /* The second and subsequent rows don't have anything in the first
     or third columns, so clear them. */
  {
    int j;
    
    for (j = prev_r + 1; j < r; j++)
      {
	som_set_null (t, 0, j);
	som_set_null (t, 2, j);
      }
  }
  
  /* Draw a line below the last row of information on this variable. */
  som_hline (t, SLIN_1, 0, 2, r);

  return r;
}

