/* *INDENT-OFF* */
!/* PSPP - computes sample statistics.
!   Copyright (C) 1997 Free Software Foundation, Inc.
!
!   This program is free software; you can redistribute it and/or
!   modify it under the terms of the GNU General Public License as
!   published by the Free Software Foundation; either version 2 of the
!   License, or (at your option) any later version.
!
!   This program is distributed in the hope that it will be useful, but
!   WITHOUT ANY WARRANTY; without even the implied warranty of
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
!   General Public License for more details.
!
!   You should have received a copy of the GNU General Public License
!   along with this program; if not, write to the Free Software
!   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!   02111-1307, USA.
!
!   The author may be contacted at <pfaffben@pilot.msu.edu> on the
!   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
!   through more mundane means. */
!
!/*
! T-TEST procedure
!*/
!

$ "t test":
$   groups=custom;
$   variables=varlist("PV_NO_SCRATCH | PV_NUMERIC");
$   *+pairs=custom;
$   +missing=miss:!analysis/listwise,
$           incl:include/!exclude;
$   +format=fmt:!labels/nolabels;
$   +criteria=:ci(n:criteria).
/* *INDENT-ON* */

/* Main declarations. */

#include <config.h>
#include <assert.h>
#include <stdio.h>

/* (headers) */

#undef DEBUGGING
#define DEBUGGING 1
#include "debug-print.h"

/* (decls global) */

/* Variable for the GROUPS subcommand, if given. */
static variable *groups;

/* GROUPS: Number of values specified by the user; the values
   specified if any. */
static int n_groups_values;
static value groups_values[2];

/* PAIRED: Number of pairs; each pair. */
static int n_pairs;
static variable *(*pairs)[2];

static int custom_groups (void);
static int custom_pairs (void);

#ifdef DEBUGGING
static void debug_print (void);
#endif

/* Parses and executes the T-TEST procedure. */
int
cmd_t_test (void)
{
  force_match_id (T);
  match_tok ('-');
  match_id (TEST);
  
  /* (parser bare) */

#if DEBUGGING
  debug_print ();
#endif

  /* Not yet implemented. */

  return 1;
}

static int parse_value (value *);

/* Parses the GROUPS subcommand. */
int
custom_groups (void)
{
  groups = parse_variable ();
  if (!groups)
    return syntax_error (_("expecting variable name in GROUPS subcommand"));
  if (groups->type == STRING && groups->width > MAX_SHORT_STRING)
    return msg (SE, _("Long string variable %s is not valid here."),
		groups->name);

  if (!match_tok ('('))
    if (groups->type == NUMERIC)
      {
	n_groups_values = 2;
	groups_values[0].f = 1;
	groups_values[1].f = 2;
	return 1;
      } else {
	return msg (SE, _("When applying GROUPS to a string variable, at "
			  "least one value must be specified."));
	return 0;
      }

  if (!parse_value (&groups_values[0]))
    return 0;
  n_groups_values = 1;

  match_tok (',');

  if (match_tok (')'))
    return 1;

  if (!parse_value (&groups_values[1]))
    return 0;
  n_groups_values = 2;

  force_match (')');

  return 1;
}

/* Parses the current token (numeric or string, depending on the
   variable in `groups') into value V and returns success. */
static int
parse_value (value *v)
{
  if (groups->type == NUMERIC)
    {
      force_num ();
      v->f = tokval;
    } else {
      force_string ();
      strncpy (v->s, tokstr, tokstr_len);
    }
  
  get_token ();
  
  return 1;
}

/* Parses the PAIRS subcommand. */
static int
custom_pairs (void)
{
  variable **vars;
  int n_before_WITH;
  int n_vars;
  int paired;
  int extra;
#if DEBUGGING
  int n_predicted;
#endif

  if ((token != ID || !is_varname (tokstr)) && token != ALL)
    return -1;
  if (!parse_variables (&default_dict, &vars, &n_vars,
			PV_DUPLICATE | PV_NUMERIC | PV_NO_SCRATCH))
    return 0;
  
  assert (n_vars);
  if (match_tok (WITH))
    {
      n_before_WITH = n_vars;

      if (!parse_variables (&default_dict, &vars, &n_vars,
			    PV_DUPLICATE | PV_APPEND
			    | PV_NUMERIC | PV_NO_SCRATCH))
	{
	  free (vars);
	  return 0;
	}
    }
  else n_before_WITH = 0;

  paired = (match_tok ('(') && match_id (PAIRED) && match_tok (')'));

  if (paired)
    {
      if (n_before_WITH * 2 != n_vars)
	{
	  free (vars);
	  return msg (SE, _("PAIRED was specified but the number of variables "
			    "preceding WITH (%d) did not match the number "
			    "following (%d)."),
		      n_before_WITH, n_vars - n_before_WITH);
	}
      
      extra = n_before_WITH;
    }
  else if (n_before_WITH)
    extra = n_before_WITH * (n_vars - n_before_WITH);
  else 
    {
      if (n_vars < 2)
	{
	  return msg (SE, _("At least two variables must be specified "
			    "on PAIRS."));
	  free (vars);
	}

      extra = n_vars * (n_vars - 1) / 2;
    }

#if DEBUGGING
  n_predicted = n_pairs + extra;
#endif

  pairs = xrealloc (pairs, sizeof (variable *[2]) * (n_pairs + extra));

  if (paired)
    {
      int i;
      
      for (i = 0; i < extra; i++)
	{
	  pairs[n_pairs][0] = vars[i];
	  pairs[n_pairs++][1] = vars[i + extra];
	}
    }
  else if (n_before_WITH)
    {
      int i;
      
      for (i = 0; i < n_before_WITH; i++)
	{
	  int j;
	  
	  for (j = n_before_WITH; j < n_vars; j++)
	    {
	      pairs[n_pairs][0] = vars[i];
	      pairs[n_pairs++][1] = vars[j];
	    }
	}
    } else {
      int i;

      for (i = 0; i < n_vars; i++)
	{
	  int j;
	  
	  for (j = i + 1; j < n_vars; j++)
	    {
	      pairs[n_pairs][0] = vars[i];
	      pairs[n_pairs++][1] = vars[j];
	    }
	}
    }

#if DEBUGGING
  assert (n_pairs == n_predicted);
#endif

  free (vars);
  return 1;
}

#if DEBUGGING
static void
debug_print (void)
{
  printf ("T-TEST\n");
  if (groups)
    {
      printf ("  GROUPS=%s", groups->name);
      if (n_groups_values)
	{
	  int i;
	  
	  printf (" (");
	  for (i = 0; i < n_groups_values; i++)
	    if (groups->type == NUMERIC)
	      printf ("%g%s", groups_values[i].f, i ? " " : "");
	    else
	      printf ("%.*s%s", groups->width, groups_values[i].s,
		      i ? " " : "");
	  printf (")");
	}
      printf ("\n");
    }
  if (n_variables)
    {
      int i;
      
      printf ("  VARIABLES=");
      for (i = 0; i < n_variables; i++)
	printf ("%s ", v_variables[i]->name);
      printf ("\n");
    }
  if (sbc_pairs)
    {
      int i;
      
      printf ("  PAIRS=");
      for (i = 0; i < n_pairs; i++)
	printf ("%s ", pairs[i][0]->name);
      printf ("WITH");
      for (i = 0; i < n_pairs; i++)
	printf (" %s", pairs[i][1]->name);
      printf (" (PAIRED)\n");
    }
  printf ("  MISSING=%s %s\n",
	  miss == ANALYSIS ? "ANALYSIS" : "LISTWISE",
	  miss == INCLUDE ? "INCLUDE" : "EXCLUDE");
  printf ("  FORMAT=%s\n",
	  fmt == LABELS ? "LABELS" : "NOLABELS");
  if (criteria != NOT_LONG)
    printf ("  CRITERIA=%ld\n", criteria);
}
#endif /* DEBUGGING */

/* Local Variables: */
/* mode:c */
/* End: */
